header {* Heurísticas para la inducción y recursion general *}

(*<*)
theory Cap_5
imports Main LaTeXsugar OptionalSugar Cap_2 Efficient_Nat
begin
(*>*)

section {* Heurísticas para la inducción *}

text {*
  \begin{definicion}[Definición recursiva de inversa]
    @{term "(inversa xs)"} es la inversa de la lista @{term "xs"}.
  \end{definicion}
*}

primrec inversa :: "'a list \<Rightarrow> 'a list" where
  "inversa [] = []"
| "inversa (x#xs) = (inversa xs) @ [x]"

text {* 
  \begin{definicion}[Definición de inversa con acumuladores]
    @{term "(inversaAc xs)"} es la inversa de la lista @{term "xs"} calculada
    con acumuladores.
  \end{definicion}
*}

primrec inversaAcAux :: "'a list \<Rightarrow> 'a list \<Rightarrow> 'a list" where
  "inversaAcAux [] ys = ys"
| "inversaAcAux (x#xs) ys = inversaAcAux xs (x#ys)"

definition inversaAc :: "'a list \<Rightarrow> 'a list" where
"inversaAc xs \<equiv> inversaAcAux xs []"

text {* 
  \begin{lema}[Ejemplo de equivalencia entre las definiciones]
    La inversa de [1,2,3] es lo mismo calculada con la primera definición que
    con la segunda.
  \end{lema}
*}

lemma "inversaAc [1,2,3] = inversa [1,2,3]"
by (simp add: inversaAc_def)

text {*
  \begin{nota}[Ejemplo fallido de demostración por inducción]
  El siguiente intento de demostrar que para cualquier lista $xs$, se tiene que
  @{term "inversaAc xs = inversa xs"} falla.
  \end{nota}
*}

lemma "inversaAc xs = inversa xs"
proof (induct xs)
  show "inversaAc [] = inversa []" by (simp add: inversaAc_def)
next
  fix a xs assume HI: "inversaAc xs = inversa xs"
  have "inversaAc (a#xs) = inversaAcAux (a#xs) []" by (simp add: inversaAc_def)
  also have "\<dots> = inversaAcAux xs [a]" by simp
  also have "\<dots> = inversa (a#xs)"
  -- "Problema: la hipótesis de inducción no es aplicable."
oops

text {* 
  \begin{nota}[Heurística de generalización]
    Cuando se use demostración estructural, cuantificar universalmente las
    variables libres (o, equivalentemente, considerar las variables libres como
    variables arbitrarias).
  \end{nota}

  \begin{lema}[Lema con generalización]
    Para toda lista $ys$ se tiene 
    \newline \hspace*{1cm} 
    @{term "inversaAcAux xs ys = (inversa xs) @ ys"}. 
  \end{lema}
*}

lemma inversaAcAux_es_inversa:
  "inversaAcAux xs ys = (inversa xs)@ys"
proof (induct xs arbitrary: ys)
  show "\<And>ys. inversaAcAux [] ys = (inversa [])@ys" by simp
next
  fix a xs 
  assume HI: "\<And>ys. inversaAcAux xs ys = inversa xs@ys"
  show "\<And>ys. inversaAcAux (a#xs) ys = inversa (a#xs)@ys"
  proof -
    fix ys
    have "inversaAcAux (a#xs) ys = inversaAcAux xs (a#ys)" by simp
    also have "\<dots> = inversa xs@(a#ys)" using HI by simp
    also have "\<dots> = inversa (a#xs)@ys" by simp 
    finally show "inversaAcAux (a#xs) ys = inversa (a#xs)@ys" by simp
  qed
qed

text {*
  \begin{corolario} Para cualquier lista $xs$, se tiene que
  @{term "inversaAc xs = inversa xs"}.
  \end{corolario}
*}

corollary "inversaAc xs = inversa xs"
by (simp add: inversaAcAux_es_inversa inversaAc_def)

text {*
  \begin{nota}
    En el paso @{term "inversa xs@(a#ys) = inversa (a#xs)@ys"} se usan
    lemas de la teoría \verb|List|. Se puede observar, activando 
    \verb|Trace Simplifier| y \verb|Trace Rules|, que los lemas usados son
    \newline \hspace*{1cm}
    \begin{tabular}[t]{ll}
    \verb|append_assoc|       & @{text "(xs @ ys) @ zs = xs @ (ys @ zs)"} \\
    \verb|append.append_Cons| & @{text "(x#xs)@ys = x#(xs@ys)"} \\
    \verb|append.append_Nil|  & @{text "[]@ys = ys"}
    \end{tabular} \newline
  Los dos últimos son las ecuaciones de la definición de \verb|append|.

  En la siguiente demostración se detallan los lemas utilizados.
  \end{nota}
*}

lemma "(inversa xs)@(a#ys) = (inversa (a#xs))@ys"
proof -
  have "(inversa xs)@(a#ys) = (inversa xs)@(a#([]@ys))" 
    by (simp only:append.append_Nil)
  also have "\<dots> = (inversa xs)@([a]@ys)" by (simp only:append.append_Cons)
  also have "\<dots> = ((inversa xs)@[a])@ys" by (simp only:append_assoc)
  also have "\<dots> = (inversa (a#xs))@ys" by (simp only:inversa.simps(2))
  finally show ?thesis .
qed

section {* Recursión general. La función de Ackermann *}

text {* 
  El objetivo de esta sección es mostrar el uso de las definiciones
  recursivas generales y sus esquemas de inducción. Como ejemplo se usa la
  función de Ackermann (se puede consultar información sobre dicha función en
  \href{http://en.wikipedia.org/wiki/Ackermann_function}{Wipipedia}).

  \begin{definicion} La función de Ackermann se define por
    $$A(m,n) =
      \begin{cases}
        n+1,             & \text{si $m=0$}, \\ 
        A(m-1,1)         & \text{si $m>0$ y $n=0$}, \\
        A(m-1,A(m,n-1)), & \text{si $m>0$ y $n>0$}
      \end{cases}$$
    para todo los números naturales. La función de Ackermann es recursiva, pero
    no es primitiva recursiva. 
  \end{definicion}
*}

fun ack :: "nat \<Rightarrow> nat \<Rightarrow> nat" where
  "ack 0 n = n+1" 
| "ack (Suc m) 0 = ack m 1" 
| "ack (Suc m) (Suc n) = ack m (ack (Suc m) n)"

text {*
  \begin{nota}[Definiciones recursivas generales]
  \mbox{}
  \begin{itemize}
  \item{Las definiciones recursivas generales se identifican mediante
    \textbf{fun}.} 
  \item{Al definir una función recursiva general se genera una regla de
    inducción. En la definición anterior, la regla generada es
    \begin{equation}\notag
      (\textit{ack.induct})\quad@{thm [mode=Rule] ack.induct}
    \end{equation}}
  \end{itemize}
  \end{nota}

  \begin{nota}[Ejemplo de cálculo]
    El cálculo del valor de la función de Ackermann para 2 y 3 se realiza
    mediante 
  \end{nota}
*}

value "ack 2 3" 

text {* 
  y se obtiene 9. 

  \begin{lema}
    Para todos $m$ y $n$, $A(m,n) > n$.
  \end{lema}
*} 

lemma "ack m n > n"
proof (induct m n rule: ack.induct)
  fix n :: "nat"
  show "ack 0 n > n" by simp
next
  fix m assume "ack m 1 > 1"
  thus "ack (Suc m) 0 > 0" by simp
next  
  fix m n
  assume "n < ack (Suc m) n" and 
         "ack (Suc m) n < ack m (ack (Suc m) n)"
  thus "Suc n < ack (Suc m) (Suc n)" by simp
qed

text {* La demostración automática es *}

lemma "ack m n > n"
by (induct m n rule: ack.induct) simp_all

text {*
  \begin{nota}[Inducción sobre recursión]
  El formato para iniciar una demostración por inducción en la regla inductiva
  correspondiente a la definición recursiva de la función $f m n$ es
  \newline \hspace*{1cm}
  \textbf{proof} (\textit{induct m n rule:f.induct}) 
  \end{nota}
*}

section {* Recursión mutua e inducción *}

text {*
  \begin{nota}[Ejemplo de definición de tipos mediante recursión cruzada]
  \mbox{}
  \begin{itemize}
  \item{Un árbol de tipo $a$ es una hoja o un nodo de tipo $a$ junto con un
  bosque de tipo $a$.}
  \item{Un bosque de tipo $a$ es el boque vacío o un bosque contruido añadiendo
  un árbol de tipo $a$ a un bosque de tipo $a$.}
  \end{itemize}
  \end{nota}
*}

datatype 'a arbol = Hoja | Nodo "'a" "'a bosque"
and      'a bosque = Vacio | ConsB "'a arbol" "'a bosque"

text {*
  \begin{nota}[Regla de inducción correspondiente a la recursión cruzada]
  La regla de inducción sobre árboles y bosques es
  (\textit{arbol\_bosque.induct}) 
  \begin{equation}
  @{thm [mode=Rule] arbol_bosque.induct } \notag
  \end{equation}
  \end{nota}

  \begin{nota}[Ejemplos de definición por recursión cruzada]
  \mbox{}
  \begin{enumerate}
  \item{\emph{(aplana\_arbol a)} es la lista obtenida aplanando el árbol 
     $a$.}   
  \item{\emph{(aplana\_bosque b)} es la lista obtenida aplanando el bosque 
     $b$.}   
  \item{\emph{(map\_arbol a h)} es el árbol obtenido aplicando la función $h$ a
     todos los nodos del árbol $a$.}   
  \item{\emph{(map\_bosque b h)} es el bosque obtenido aplicando la función $h$
     a todos los nodos del bosque $b$.}   
  \end{enumerate}
  \end{nota}
*}

fun
  aplana_arbol :: "'a arbol \<Rightarrow> 'a list" and 
  aplana_bosque :: "'a bosque \<Rightarrow> 'a list" where
  "aplana_arbol Hoja = []"
| "aplana_arbol (Nodo x b) = x#(aplana_bosque b)"
| "aplana_bosque Vacio = []"
| "aplana_bosque (ConsB a b) = (aplana_arbol a) @ (aplana_bosque b)"

fun
  map_arbol :: "'a arbol \<Rightarrow> ('a \<Rightarrow> 'b) \<Rightarrow> 'b arbol" and
  map_bosque :: "'a bosque \<Rightarrow> ('a \<Rightarrow> 'b) \<Rightarrow> 'b bosque" where
  "map_arbol Hoja h = Hoja"
| "map_arbol (Nodo x b) h = Nodo (h x) (map_bosque b h)"
| "map_bosque Vacio h = Vacio"
| "map_bosque (ConsB a b) h = ConsB (map_arbol a h) (map_bosque b h)"

text {*
  \begin{lema}[Ejemplo de inducción cruzada]
  \mbox{}
  \begin{enumerate}
    \item{@{term "aplana_arbol (map_arbol a h) = map h (aplana_arbol a)"}}
    \item{@{term "aplana_bosque (map_bosque b h) = map h (aplana_bosque b)"}}
  \end{enumerate}
  \end{lema}
*}

lemma "aplana_arbol (map_arbol a h) = map h (aplana_arbol a)
     \<and> aplana_bosque (map_bosque b h) = map h (aplana_bosque b)"
proof (induct_tac a and b)
  show "aplana_arbol (map_arbol Hoja h) = map h (aplana_arbol Hoja)" by simp
next
  fix x b
  assume HI: "aplana_bosque (map_bosque b h) = map h (aplana_bosque b)"
  have "aplana_arbol (map_arbol (Nodo x b) h) 
        = aplana_arbol (Nodo (h x) (map_bosque b h))" by simp
  also have "\<dots> = (h x)#(aplana_bosque (map_bosque b h))" by simp
  also have "\<dots> = (h x)#(map h (aplana_bosque b))" using HI by simp
  also have "\<dots> = map h (aplana_arbol (Nodo x b))" by simp
  finally show "aplana_arbol (map_arbol (Nodo x b) h)
                = map h (aplana_arbol (Nodo x b))" .
next
  show "aplana_bosque (map_bosque Vacio h) = map h (aplana_bosque Vacio)" by simp
next
  fix a b
  assume HI1: "aplana_arbol (map_arbol a h) = map h (aplana_arbol a)"
     and HI2: "aplana_bosque (map_bosque b h) = map h (aplana_bosque b)"
  have "aplana_bosque (map_bosque (ConsB a b) h)
        = aplana_bosque (ConsB (map_arbol a h) (map_bosque b h))" by simp
  also have "\<dots> = aplana_arbol(map_arbol a h)@aplana_bosque(map_bosque b h)" 
    by simp
  also have "\<dots> = (map h (aplana_arbol a))@(map h (aplana_bosque b))" 
    using HI1 HI2 by simp
  also have "\<dots> = map h (aplana_bosque (ConsB a b))" by simp
  finally show "aplana_bosque (map_bosque (ConsB a b) h) 
                = map h (aplana_bosque (ConsB a b))" by simp
qed

lemma "aplana_arbol (map_arbol a h) = map h (aplana_arbol a)
     \<and> aplana_bosque (map_bosque b h) = map h (aplana_bosque b)"
by (induct_tac a and b) auto

(*<*)
end
(*>*)

